# Copyright (C) 2015-2020 Guillaume VINET
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

from pynfcreader.devices.devices import Devices
from pyHydrabus import NFC


class HydraNFC(Devices):

    def __init__(self, port="C0M8", debug=True):

        self._port = port
        self._hydranfc = None

        self.__logger = logging.getLogger()
        stream_handler = logging.StreamHandler()
        stream_debug_formatter = logging.Formatter('%(levelname)s  ::  %(message)s')
        stream_handler.setFormatter(stream_debug_formatter)

        self.__logger.setLevel(logging.INFO)
        stream_handler.setLevel(logging.INFO)

        self._python_ver = sys.version[0]

        if debug:
            self.__logger.setLevel(logging.DEBUG or logging.INFO)
            stream_handler.setLevel(logging.DEBUG or logging.INFO)

        self.__logger.addHandler(stream_handler)

    def connect(self):
        self.__logger.info("Connect to HydraNFC")
        self.__logger.info("")
        self._hydranfc = NFC(self._port)

    def get_logger(self):
        return self.__logger

    def set_mode_iso14443A(self):
        self._hydranfc.mode = self._hydranfc.MODE_ISO_14443A

    def set_mode_iso15693(self):
        self._hydranfc.mode = self._hydranfc.MODE_ISO_15693

    def field_off(self):
        self.__logger.debug("Field off")
        self._hydranfc.rf = 0

    def field_on(self):
        self.__logger.debug("Field on")
        self._hydranfc.rf = 1

    def write_bits(self, data=b"", num_bits=0):
        resp = self._hydranfc.write_bits(data, num_bits)
        self.__logger.debug(f"\t<{' '.join(f'{hit:02X}' for hit in resp)}")
        self.__logger.debug("")
        return resp

    def write(self, data=b"", resp_len=None, transmitter_add_crc=True):
        self.__logger.debug("write")
        self.__logger.debug(f"\t>{data.hex()}")

        resp = self._hydranfc.write(data, transmitter_add_crc)
        self.__logger.debug(f"\t<{data.hex()}")
        self.__logger.debug("")

        return resp



# Copyright (C) 2015-2020 Guillaume VINET
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

from pynfcreader.devices.devices import Devices
import serial

class HydraNFCv2(Devices):

    def __init__(self, port="C0M8", debug=True):

        self._port = port
        self._hydranfc = None

        self.__logger = logging.getLogger()
        stream_handler = logging.StreamHandler()
        stream_debug_formatter = logging.Formatter('%(levelname)s  ::  %(message)s')
        stream_handler.setFormatter(stream_debug_formatter)

        self.__logger.setLevel(logging.INFO)
        stream_handler.setLevel(logging.INFO)

        self._python_ver = sys.version[0]

        if debug:
            self.__logger.setLevel(logging.DEBUG or logging.INFO)
            stream_handler.setLevel(logging.DEBUG or logging.INFO)

        self.__logger.addHandler(stream_handler)

    def enter_bbio(self):
        self._hydranfc.timeout = 0.01
        for _ in range(20):
            self._hydranfc.write(b"\x00")
            if b"BBIO1" in self._hydranfc.read(5):
                self._hydranfc.reset_input_buffer()
                self._hydranfc.timeout = None

                # We enter reader mode
                self._hydranfc.write(b"\x0E")
                if self._hydranfc.read(4) != b"NFC2":
                    raise Exception("Cannot enter BBIO Reader mode")
                return
        raise Exception("Cannot enter BBIO mode.")

    def connect(self):
        self.__logger.info("Connect to HydraNFC")
        self.__logger.info("")
        self._hydranfc = serial.Serial(self._port, timeout=None)
        self.enter_bbio()

    def get_logger(self):
        return self.__logger

    def set_mode_iso14443A(self):
        self._hydranfc.write(b"\x06")

    def set_mode_iso14443B(self):
        self._hydranfc.write(b"\x09")

    def set_mode_iso15693(self):
        self._hydranfc.write(b"\x07")

    def field_off(self):
        self.__logger.debug("Field off")
        self._hydranfc.write(b"\x02")

    def field_on(self):
        self.__logger.debug("Field on")
        self._hydranfc.write(b"\x03")

    def write_bits(self, data=b"", num_bits=0):
        # resp = self._hydranfc.write_bits(data, num_bits)

        self._hydranfc.write(b"\x08")
        rx_len = int.from_bytes(self._hydranfc.read(1), byteorder="little")
        resp  = self._hydranfc.read(rx_len)

        self.__logger.debug(f"\t<{' '.join(f'{hit:02X}' for hit in resp)}")
        self.__logger.debug("")

        return resp

    def write(self, data=b"", resp_len=None, transmitter_add_crc=True):
        self.__logger.debug("write")
        self.__logger.debug(f"\t>{data.hex()}")

        self._hydranfc.write(b"\x05")
        self._hydranfc.write(int(transmitter_add_crc).to_bytes(1, byteorder="big"))
        self._hydranfc.write(len(data).to_bytes(1, byteorder="big"))
        self._hydranfc.write(data)

        rx_len = int.from_bytes(self._hydranfc.read(1), byteorder="little")

        resp =  self._hydranfc.read(rx_len)

        self.__logger.debug(f"\t<{data.hex()}")
        self.__logger.debug("")

        return resp



# Copyright (C) 2015-2020 Guillaume VINET
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from pynfcreader.devices.hydra_nfc_v2 import HydraNFCv2
from pynfcreader.sessions.iso14443.iso14443b import Iso14443BSession

hydra_nfc = HydraNFCv2(port="/dev/ttyACM0", debug=False)
hn = Iso14443BSession(drv=hydra_nfc, block_size=120)

hn.connect()
hn.field_off()
time.sleep(0.1)
hn.field_on()
hn.polling()
print(hn.pupi)

r = hn.send_apdu("00 a4 04 00   0E   32 50 41 59 2E 53 59 53 2E 44 44 46 30 31   00")
print(r)
r =  hn.send_apdu("00 a4 04 00   07   A0 00 00 00 42 10 10  00")
print(r)
r = hn.send_apdu("00 a4 04 00   07   A0 00 00 00 04 10 10   00")
print(r)

hn.field_off()



# Copyright (C) 2015-2020 Guillaume VINET
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pynfcreader.sessions.iso15693.requests import \
    RequestInventory, \
    RequestReadSingleBlock, \
    RequestReadMultipleBlocks, \
    RequestGetSystemInformation, \
    RequestStayQuiet, \
    RequestWriteSingleBlock, \
    RequestWriteMultipleBlock, \
    RequestSelect, \
    RequestGetMultipleBlockSecurityStatus
from pynfcreader.tools import utils


class Iso15693Session(object):

    def __init__(self, drv=None):
        self._drv = drv
        self._logger = self._drv.get_logger()
        self.last_request = None
        self._memory_block = {}
        self._lock_status = {}

    def connect(self):
        self._drv.connect()
        self._drv.set_mode_iso15693()

    def field_on(self):
        self._logger.info("field on")
        self._drv.field_on()
        self._logger.info("")

    def field_off(self):
        self._logger.info("field off")
        self._drv.field_off()
        self._logger.info("")

    def send(self, cmd):

        return self._drv.write(data=cmd, resp_len=16, transmitter_add_crc=True)

    def send_cmd(self, cmd, no_answer=False):
        self._logger.info(f"Command {cmd.name}")

        for hit in utils.get_pretty_print_block(cmd()):
            self._logger.info(f"{hit}")

        for key, value in cmd.items.items():
            if value != b"":
                self._logger.info(f"\t{key:20}: {utils.bytes_to_str(value)}")
        self._logger.info("")

        resp = self.send(cmd())

        self.last_request = cmd

        if no_answer:
            return

        self._logger.info("Response:")
        if resp:
            for hit in utils.get_pretty_print_block(resp):
                self._logger.info(hit)
        else:
            self._logger.info("")
        try:
            cmd.resp_pretty_print(resp)
        except:  # noqa: E722
            pass
        for key, value in cmd.resp.items():
            if value != b"":
                self._logger.info(
                    f"\t{key:25}: {utils.bytes_to_str(value['raw'])}")
                if value["pretty"] != "":
                    self._logger.info(f"\t{' ' * 25}: {value['pretty']}")
        self._logger.info("")

        return resp

    def inventory(self, flags=b"\x26", afi_opt=b"", mask=b""):
        return self.send_cmd(RequestInventory(flags, afi_opt, mask))

    def stay_quiet(self, flags=b"\x22", uid=b""):
        return self.send_cmd(RequestStayQuiet(flags, uid), no_answer=True)

    def read_single_block(self, flags=b"\x42", uid_opt=b"", block_nb=b""):
        return self.send_cmd(RequestReadSingleBlock(flags, uid_opt, block_nb))

    def write_single_block(self, flags=b"\x42", uid_opt=b"", block_nb=b"",
                           data=b""):
        return self.send_cmd(
            RequestWriteSingleBlock(flags, uid_opt, block_nb, data))

    def get_system_info(self, flags=b"\x22", uid_opt=b""):
        return self.send_cmd(RequestGetSystemInformation(flags, uid_opt))

    def read_multiple_blocks(self, flags=b"\x42", uid_opt=b"",
                             first_block_nb=b"\x00", nb_blocks=b"\x00"):
        return self.send_cmd(
            RequestReadMultipleBlocks(flags, uid_opt, first_block_nb,
                                      nb_blocks))

    def write_multiple_block(self, flags=b"\x02", uid_opt=b"",
                             first_block_nb=b"\x00", nb_blocks=b"\x00",
                             data=b""):
        return self.send_cmd(
            RequestWriteMultipleBlock(flags, uid_opt, first_block_nb, nb_blocks,
                                      data))

    def select(self, flags=b"\x22", uid=b""):
        return self.send_cmd(RequestSelect(flags, uid))

    def get_multiple_blocks_security_status(self, flags=b"\x02", uid_opt=b"",
                                            first_block_nb=b"\x00",
                                            nb_blocks=b"\x00"):
        return self.send_cmd(
            RequestGetMultipleBlockSecurityStatus(flags, uid_opt,
                                                  first_block_nb, nb_blocks))

    def get_all_auto(self):
        uid = self.inventory()[2:][::-1]
        self.get_system_info(uid_opt=uid)
        nb_block = self.last_request.nb_block
        self.get_all_memory_info(nb_block)

        self._logger.info("\tMemory dump")
        self._logger.info("")

        for hit in range(nb_block):
            block = utils.int_array_to_hex_str(self._memory_block[hit])
            block_ascii = utils.bytes_to_ascii_printable_str(
                self._memory_block[hit])
            status = "Locked  " if (
                    self._lock_status[hit][0] & 0x1) else "Unlocked"
            self._logger.info(
                f"\t\t[{hit:3d}] - {status} -  {block} | {block_ascii}")

    def get_all_memory_info(self, block_num):
        self._logger.info("Get and print all memory")

        self._memory_block = {}
        self._lock_status = {}

        self._logger.info("\tGet all memory")
        for hit in range(block_num):
            block_nb = bytes([hit])
            resp = self.read_single_block(block_nb=block_nb)[2:]
            if len(resp) == 0:
                self.read_single_block(block_nb=block_nb)[2:]
            self._memory_block[hit] = self.last_request.resp["data"]["raw"]
            self._lock_status[hit] = \
                self.last_request.resp["block_security_status"]["raw"]



import time

from pynfcreader.sessions.iso14443.iso14443b import Iso14443BSession


def test_iso_14443_a_card_1_generic(hydranfc_connection):
    hn = Iso14443BSession(drv=hydranfc_connection, block_size=120)

    hn.connect()
    hn.field_off()
    time.sleep(0.1)
    hn.field_on()
    hn.polling()

    r = hn.send_apdu("00 a4 04 00   0E   32 50 41 59 2E 53 59 53 2E 44 44 46 30 31   00")
    assert b'o[\x84\x0e2PAY.S.DDF01\xa5I\xbf\x0cFO\x07\xa0\x00\x00\x00B\x10\x10P\x02\x87\x01\x01\x9f*\x01\x02\x9f\n\x08\x00\x01\x00\x00\x00\x00a#O\x07\xa0\x00\x04\x10\x10P\nMASTERRD\x87\x01\x01\x9f\n\x08\x00\x01\x05\x00\x00\x00' == r

    hn.field_off()



# Copyright (C) 2015-2020 Guillaume VINET
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re

manufacturer_codes_iso_7816_6 = {
    "01": "Motorola",
    "02": "ST Microelectronics",
    "03": "Hitachi",
    "04": "NXP Semiconductors",
    "05": "Infineon Technologies",
    "06": "Cylinc",
    "07": "Texas Instruments Tag-it",
    "08": "Fujitsu Limited",
    "09": "Matsushita Electric Industrial",
    "0A": "NEC",
    "0B": "Oki Electric",
    "0C": "Toshiba",
    "0D": "Mitsubishi Electric",
    "0E": "Samsung Electronics",
    "0F": "Hyundai Electronics",
    "10": "LG Semiconductors",
    "16": "EM Microelectronic-Marin",
    "1F": "Melexis",
    "2B": "Maxim",
    "33": "AMIC"}


def int_array_to_hex_str(array):
    return " ".join(f"{hit:02X}" for hit in array)


def bytes_to_ascii_printable_str(data):
    data = data.decode(encoding='ascii', errors='replace')
    return re.sub(r'[^\x20-\x7E]', '.', data)


def bytes_to_str(data_b):
    return ' '.join(f'{hit:02X}' for hit in data_b)


def get_pretty_print_block(msg):
    lst = []

    for hit in range(0, len(msg), 16):
        block1 = msg[hit:hit + 8]
        hex1 = " ".join(f"{c:02X}" for c in block1)
        str1 = bytes_to_ascii_printable_str(block1)
        block2 = msg[hit + 8:hit + 16]
        hex2 = " ".join(f"{c:02X}" for c in block2)
        str2 = bytes_to_ascii_printable_str(block2)

        lpart = f"{hex1}   {hex2}"
        lst.append(f'{lpart}{" " * (49 - len(lpart))}      {str1}   {str2}')
    return lst



# Copyright (C) 2015-2020 Guillaume VINET
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pynfcreader.sessions.iso14443a.tpdu import Tpdu
from pynfcreader.tools import utils


class Iso14443ASession(object):

    def __init__(self, cid=0, nad=0, drv=None, block_size=16):
        self._init_pcb_block_nb()
        self._addNAD = False
        self._addCID = True
        self._cid = cid
        self._nad = nad
        self._drv = drv
        self._iblock_pcb_number = 0x00
        self._pcb_block_number = None
        self._drv = drv
        self._logger = self._drv.get_logger()
        self.block_size = block_size

    def connect(self):
        self._drv.connect()
        self._drv.set_mode_iso14443A()

    def field_on(self):
        self._logger.info("field on")
        self._drv.field_on()

    def field_off(self):
        self._logger.info("field off")
        self._drv.field_off()

    def polling(self):
        self.send_reqa()
        self.send_select_full()
        self.send_pps()

    @property
    def block_size(self):
        return self._block_size

    @block_size.setter
    def block_size(self, size):
        assert (0 <= size <= 256)
        self._block_size = size

    def get_and_update_iblock_pcb_number(self):
        self._iblock_pcb_number ^= 1
        return self._iblock_pcb_number ^ 1

    def send_reqa(self):
        """
        REQA = REQ frame - Type A
        0x26 - 7 bits - no CRC

        ret :
          - nothing
          - ATQA (Answer To Request - Type A)
        """
        self.comment_data("REQA (7 bits):", b"\x26")
        resp = self._drv.write_bits(b'\x26', 7)
        if not resp:
            raise Exception("REQ A failure")
        self.comment_data("ATQA:", resp)
        return resp

    def send_wupa(self):
        """
        REQA = REQ frame - Type A
        0x26 - 7 bits - no CRC

        ret :
          - nothing
          - ATQA (Answer To Request - Type A)
        """
        self.comment_data("WUPA (7 bits):", [0x52])
        resp = self._drv.write_bits(b'\x52', 7)
        self.comment_data("ATQA:", resp)

    def send_select_full(self, FSDI="0", CID="0"):
        """
        Select
        0x9320 - 8 bits - no CRC
        """

        uid1 = bytes()
        uid2 = bytes()
        uid3 = bytes()

        # 93 : Select cascade level 1
        # 20 : 2 * 8 + 0 = 16 bits = 2 bytes transmitted
        # No CRC
        data = bytes([0x93, 0x20])
        self.comment_data("Select cascade level 1:", data)
        resp = self._drv.write(data=data, transmitter_add_crc=False)
        self.comment_data("Select cascade level 1 response:", resp)

        # resp : CT + UID (3 bytes) + BCC
        # 93 : Select cascade level 1
        # 70 : 7 * 8 + 0 = 56 bits  = 7 bytes transmitted
        # 4 bytes : CT + UID
        # CRC_A (2 bytes)
        uid1 = resp
        data = bytes([0x93, 0x70]) + uid1
        self.comment_data("Select cascade level 1:", data)
        resp = self._drv.write(data=data, resp_len=3, transmitter_add_crc=True)
        self.comment_data("Select cascade level 1 response:", resp)

        # if uid1[0] == 0x88:
        #
        #     # resp : SAK + CRC_A
        #     # Cascade bit set? => UID not complete
        #     if resp[0] & 0x04:
        #         # 95 : Select cascade level 2 (SEL)
        #         # 20 : 2 * 8 + 0 = 16 bits = 2 bytes transmitted
        #         # No CRC
        #         data = [0x95, 0x20]
        #         self.comment_data("Select cascade level 2:", data)
        #         resp = self.__drv.write(data = data, resp_len = 5)
        #         self.comment_data("Select cascade level 2 response:", resp)
        #
        #         # resp : 4 bytes (UID) (+BCC not see in this implementation)
        #         # 95 : Select cascade level 2 (SEL)
        #         # 70 : 7 * 8 + 0 = 56 bits  = 7 bytes transmitted
        #         # No CRC
        #         uid2 = resp
        #         data = [0x95, 0x70] + uid2
        #         self.comment_data("Select cascade level 2:", data)
        #         resp = self.__drv.write(data = data, resp_len = 3, crc_in_cmd=False)
        #         self.comment_data("Select cascade level 2 response:", resp)
        #
        #         # resp : SAK + CRC_A
        #         # Cascade bit set? => UID not complete
        #         if resp[0] & 0x04:
        #             # 97 : Select cascade level 3 (SEL)
        #             # 20 : 2 * 8 + 0 = 16 bits = 2 bytes transmitted
        #             # No CRC
        #             # resp = self.hf14a_raw(data = "9720", leaveSigOnAfterRec= True, fieldMode = "OnWithoutSelect")
        #             self.__logger.info("Select cascade level 3")
        #             resp = self.__drv.write(data = [0x97, 0x20], resp_len = 3)
        #
        #             # resp : 4 bytes (UID) + BCC
        #             # 97 : Select cascade level 32 (SEL)
        #             # 70 : 7 * 8 + 0 = 56 bits  = 7 bytes transmitted
        #             # No CRC
        #             uid2 = resp
        #             # resp = self.hf14a_raw(data = "9770" + resp, leaveSigOnAfterRec= True, fieldMode = "OnWithoutSelect", AutoAddCRC = True)
        #             self.__logger.info("Select cascade level 3")
        #             resp = self.__drv.write(data = [0x97, 0x70] + resp, resp_len = 3, crc_in_cmd=False)
        #
        # RATS (Request Answer To Select)
        resp = self.send_rats_a(FSDI, CID)
        return (uid1 + uid2 + uid3, resp)

    def send_rats_a(self, FSDI="0", CID="0"):
        """
        Request for answer to select - Type A
        0xE0 - FSDI | CID - CRC_A
        :param FSDI: defines the maximum size of a frame the PCD is able to receive.
            - 0 : 16 bytes
            - 1 : 24 bytes
            - 2 : 32
            - 3 : 40
            - 4 : 48
            - 5 : 64
            - 6 : 96
            - 7 : 128
            - 8 : 256
            - .. : RFU
        :param CID: logical number of the addressed PICC in the range from 0 to 14.
        :return: ATS (Answer to select)
        """
        dico = {"0": 16, "1": 24, "2": 32, "3": 40, "4": 48, "5": 64, "6": 96, "7": 128, "8": 256}
        self._logger.info("Request for Answer To Select (RATS):")
        self._logger.info("\tPCD selected options:")
        self._logger.info("\t\tFSDI : 0x%s => max PCD frame size : %d bytes" % (FSDI, dico[FSDI]))
        self._logger.info("\t\tCID  : 0x%s" % FSDI)

        data = bytes([0xE0, int(FSDI + CID, 16)])
        self.comment_data("RATS", data)
        resp = self._drv.write(data=data, resp_len=20, transmitter_add_crc=True)

        # resp[0] = TL = length without counting the 2 CRC bytes
        resp = resp[:resp[0] + 2]
        self.comment_data("Answer to Select (ATS = RATS response):", resp)

        T0 = resp[1]
        self._logger.info("\tT0 : 0x%02X", T0)
        self._logger.info("\t\tFSCI : 0x%01X => max card frame size : %d bytes" % (T0 & 0xF, dico[str(T0 & 0xF)]))
        ta1 = None
        tb1 = None
        tc1 = None
        cmpt = 2
        if T0 & 0x10:
            self._logger.info("\t\tTA(1) present")
            ta1 = resp[cmpt]
            cmpt += 1
        if T0 & 0x20:
            self._logger.info("\t\tTB(1) present")
            tb1 = resp[cmpt]
            cmpt += 1
        if T0 & 0x40:
            self._logger.info("\t\tTC(1) present")
            tc1 = resp[cmpt]
            cmpt += 1

        if ta1:
            self._logger.info("\tTA(1) : 0x%02X" % ta1)
            self._logger.info("\t\tInterpretation : TODO...")
        if tb1:
            self._logger.info("\tTB(1) : 0x%02X" % tb1)
            self._logger.info("\t\tInterpretation : TODO...")
        if tc1:
            self._logger.info("\tTC(1) : 0x%02X" % tc1)
            self._addNAD = ((tc1 & 0x01) == 0x01)
            self._logger.info("\t\tNAD %ssupported" % ((not self._addNAD) * "not "))
            self._addCID = ((tc1 & 0x02) == 0x02)
            self._logger.info("\t\tCID %ssupported" % ((not self._addCID) * "not "))
        self._logger.info("\tHistorical bytes : %s" % utils.int_array_to_hex_str(resp[cmpt:-2]))
        self._logger.info("\tCRC : %s" % utils.int_array_to_hex_str(resp[-2:]))

        self._logger.info("")
        self._logger.info("")
        self._logger.info("")
        return resp

    def send_pps(self, cid=0x0, pps1=False, dri=0x0, dsi=0x0):
        self._logger.info("PPS")
        self._logger.info("\tPCD selected options:")
        pps0_pps1 = [0x01]
        if pps1:
            pps0_pps1.append(dri << 4 + dsi)
        self._logger.info("\tCID : 0x%X" % cid)
        self._logger.info("\tPPS1 %stransmitted" % ("not " * (not pps1)))

        data = bytes([0xD0 + cid] + pps0_pps1)
        self.comment_data("PPS:", data)
        resp = self._drv.write(data=data, resp_len=3, transmitter_add_crc=True)
        self.comment_data("PPS response:", resp)
        if resp[0] == (0xD0 + cid):
            self._logger.info("\tPPS accepted")
        else:
            self._logger.info("\tPPS rejected")
        self._logger.info("")

        return resp

    def _init_pcb_block_nb(self):
        self._pcb_block_number = 0

    def _inc_pcb_block_number(self):
        self._pcb_block_number ^= 1

    def _get_pcb_block_number(self):
        return self._pcb_block_number

    def _get_and_update_pcb_block_number(self):
        nb = self._get_pcb_block_number()
        self._inc_pcb_block_number()
        return nb

    def get_rblock(self, ack=True, cid=True, block_number=None):

        pcb = 0xA2
        cid = ""
        if not ack:
            pcb |= 0x10

        if cid:
            pcb |= 0x08
            cid = f" 0x{self._cid:02X} "

        if block_number:
            pcb |= block_number
        else:
            pcb |= self.get_and_update_iblock_pcb_number()

        if cid != "":
            return bytes([pcb, cid])
        else:
            return bytes([pcb])

    def getIBlock(self, data, chaining_bit=False):
        """
         - 0
         - 0
         - 0
         - Chaining if 1
         - CID following if 1
         - NAD following if 1
         - 1
         - Block number
        """
        pcb = self.get_and_update_iblock_pcb_number() + 0x02

        cid = ""
        if self._addCID:
            cid = self._cid
            pcb |= 0x08

        if chaining_bit:
            pcb |= 0x10

        nad = ""
        if self._addNAD:
            nad = self._nad
            pcb |= 0x04

        header = [pcb]
        if nad != "":
            header.append(nad)
        if cid != "":
            header.append(cid)

        return bytes(header) + data

    def chaining_iblock(self, data=None, block_size=None):

        if not block_size:
            block_size = self.block_size

        block_lst = []
        fragmented_data_index = range(0, len(data), block_size)
        for hit in fragmented_data_index[:-1]:
            inf_field = data[hit:hit + block_size]
            frame = self.getIBlock(inf_field, chaining_bit=True)
            block_lst.append(frame)

        if fragmented_data_index[-1]:
            index = fragmented_data_index[-1]
        else:
            index = 0
        inf_field = data[index:index + block_size]
        frame = self.getIBlock(inf_field, chaining_bit=False)
        block_lst.append(frame)

        return block_lst

    def _send_tpdu(self, tpdu):
        self._logger.info("\t\t" + "TPDU command:")
        for hit in utils.get_pretty_print_block(tpdu):
            self._logger.info("\t\t" + hit)

        resp = self._drv.write(data=tpdu, resp_len=16, transmitter_add_crc=True)

        resp = Tpdu(resp)
        self._logger.info("\t\t" + "TPDU response:")
        for hit in utils.get_pretty_print_block(resp.get_tpdu()):
            self._logger.info("\t\t" + hit)
        return resp

    def send_apdu(self, apdu):
        apdu = bytes.fromhex(apdu)
        self._logger.info("APDU command:")
        for hit in utils.get_pretty_print_block(apdu):
            self._logger.info("\t" + hit)

        block_lst = self.chaining_iblock(data=apdu)

        if len(block_lst) == 1:
            resp = self._send_tpdu(block_lst[0])
        else:
            self._logger.info("Block chaining, %d blocks to send" % len(block_lst))
            for iblock in block_lst:
                resp = self._send_tpdu(iblock)

        while resp.is_wtx():
            wtx_reply = resp.get_wtx_reply()
            resp = self._send_tpdu(wtx_reply)

        rapdu = resp.get_inf_field()

        while resp.is_chaining():
            rblock = self.get_rblock()

            resp = self._send_tpdu(rblock)

            rapdu += resp.get_inf_field()

        self._logger.info("APDU response:")
        for hit in utils.get_pretty_print_block(rapdu):
            self._logger.info("\t" + hit)
        return rapdu

    def comment_data(self, msg, data):
        self._logger.info(msg)
        for hit in utils.get_pretty_print_block(data):
            self._logger.info("\t" + hit)



# Copyright (C) 2015-2020 Guillaume VINET
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time
from pynfcreader.devices.hydra_nfc_v2 import HydraNFCv2
from pynfcreader.sessions.iso15693.iso15693 import Iso15693Session

hydra_nfc = HydraNFCv2(port="/dev/ttyACM1", debug=False)

hn = Iso15693Session(drv=hydra_nfc)
hn.connect()
hn.field_off()
time.sleep(0.2)
hn.field_on()
hn.inventory()
hn.get_all_auto()
# uid = hn.inventory()[2:][::-1]
# hn.get_system_info(uid_opt=uid)
hn.write_single_block(block_nb=bytes([3]), data=bytes.fromhex("AA" * 4))



# Copyright (C) 2015-2020 Guillaume VINET
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from pynfcreader.tools.utils import manufacturer_codes_iso_7816_6


class Request(object):
    cmd = {b"\x01": "Inventory",
           b"\x02": "Stay quiet",
           b"\x20": "Read single block",
           b"\x21": "Write single block",
           b"\x23": "Read multiple blocks",
           b"\x24": "Write multiple blocks",
           b"\x25": "Select",
           b"\x26": "Reset to ready",
           b"\x2B": "Get system information",
           b"\x2C": "Get multiple block security status"}

    error_code_def = {1: "The command is not supported, i.e. the request code is not recognised.",
                      2: "The command is not recognised, for example: a format error occurred.",
                      3: "The option is not supported.",
                      4: "Unknown error.",
                      5: "The specified block is not available (doesn’t exist).",
                      6: "The specified block is already -locked and thus cannot be locked again",
                      0xF: "Unknown error.",
                      0x10: "The specified block is not available (doesn’t exist).",
                      0x11: "The specified block is already -locked and thus cannot be locked again",
                      0x12: "The specified block is locked and its content cannot be changed.",
                      0x13: "The specified block was not successfully programmed.",
                      0x14: "The specified block was not successfully locked."}

    def __init__(self, **kwargs):

        self.items = collections.OrderedDict(kwargs)
        self.name = self.cmd[kwargs["command"]]
        self.resp = collections.OrderedDict()

    def get_error_code(self, error_code):
        if error_code == 0:
            return "No error"

        if error_code in self.error_code_def.keys():
            return self.error_code_def[error_code]

        if error_code in range(0xA0, 0xE0):
            return "Custom command error codes"

        return "DFU"

    def get_pretty_security_status(self, status):
        return 'Locked' if status[0] == 1 else 'Unlocked'

    def pretty_print_request_info_flags(self, info_flags):
        info_flags_pretty = collections.OrderedDict({"DSFID": {0: "DSFID is not supported. DSFID field is not present",
                                                               1: "DSFID is supported. DSFID field is present"},
                                                     "AFI": {0: "AFI is not supported. AFI field is not present",
                                                             1: "AFI is supported. AFI field is present"},
                                                     "VICC memory size": {
                                                         0: "Information on VICC memory s ize is not supported. Memory size field is not present.",
                                                         1: "Information on VICC memory s ize is supported. Memory size field is present."},
                                                     "IC reference": {0: "Information on IC reference is not supported. IC reference field is not present.",
                                                                      1: "Information on IC reference is supported. IC reference field is present."}})

        pretty = ""
        bit = 0
        for key in info_flags_pretty.keys():
            pretty += f"{info_flags_pretty[key][(info_flags >> bit) & 1]}\n"
        return pretty

    def pretty_print_request_flags(self, flag):
        info = collections.OrderedDict()
        request_flags_b1_to_b4 = {"Sub-carrier_flag": {0: "A single sub-carrier frequency shall be used by the VICC",
                                                       1: "Two sub-carriers shall be used by the VICC"},
                                  "Data_rate_flag": {0: "Low data rate shall be used",
                                                     1: "High data rate shall be used"},
                                  "Inventory_flag": {0: "Inventory flag off",
                                                     1: "Inventory flag on"},
                                  "Protocol_Extension_flag": {0: "No protocol format extension",
                                                              1: "Protocol format is extended. Reserved for future use"}}

        request_flags_b5_to_b8_inventory = {"AFI_flag": {0: "AFI field is not present",
                                                         1: "AFI field is present"},
                                            "Nb_slots_flag": {0: "16 slots",
                                                              1: "1 slots"},
                                            "Option_flag": {
                                                0: "Meaning is defined by the command description. It shall be set to 0 if not otherwise defined by the command.",
                                                1: "Meaning is defined by the command description."},
                                            "RFU": {0: "Shall be set to 0.",
                                                    1: "Shall be set to 0."}}

        request_flags_b5_to_b8_no_inventory = {"Select_flag": {0: "Request shall be executed by any VICC according to the setting of Address_flag",
                                                               1: "Request shall be executed only by VICC in selected state"},
                                               "Address_flag": {0: "Request is not addressed. UID field is not present. It shall be executed by any VICC.",
                                                                1: "Request is addressed. UID field is present. It shall be executed only by the VICC whose UID matches the UID specified in the request."},
                                               "Option_flag": {
                                                   0: "Meaning is defined by the command description. It shall be set to 0 if not otherwise defined by the command.",
                                                   1: "Meaning is defined by the command description."},
                                               "RFU": {0: "Shall be set to 0.",
                                                       1: "Shall be set to 0."}}

        # for bit in range(4):
        #     print(f"{key:20}: {' '.join(f'{hit:02X}' for hit in value)}")

    def cmd_pretty_print(self):
        for key, value in self.items.items():
            if value != b"":
                print(f"{key:20}: {' '.join(f'{hit:02X}' for hit in value)}")

    def resp_pretty_print(self, resp=b""):
        flags = resp[0:1]
        self.resp["flags"] = {"raw": flags,
                              "pretty": f"{self.get_error_code(flags)} ({flags:02X})"}

        # Error flag?
        if resp[0] & 1:
            error_code = resp[1:2]
            self.resp["error_code"] = {"raw": error_code,
                                       "pretty": f"{self.get_error_code(error_code)} ({error_code:02X})"}

    def __call__(self):
        return b"".join(hit for hit in self.items.values() if hit != b"")


class RequestInventory(Request):

    def __init__(self, flags=b"\x00", afi_opt=b"", mask_opt=b""):
        Request.__init__(self,
                         flags=flags,
                         command=b"\x01",
                         afi_opt=afi_opt,
                         mask_len=bytes([len(mask_opt)]),
                         mask_opt=mask_opt)

    def resp_pretty_print(self, resp=b""):
        flags = resp[0:1]
        self.resp["flags"] = {"raw": flags,
                              "pretty": f"0x{flags[0]:02X}"}
        dsfid = resp[1:2]
        self.resp["dsfid"] = {"raw": dsfid,
                              "pretty": f"0x{dsfid[0]:02X}"}

        uid = resp[2:10]
        self.resp["uid"] = {"raw": uid,
                            "pretty": " ".join(f"{hit:02X}" for hit in resp[2:10][::-1])}


class RequestStayQuiet(Request):

    def __init__(self,
                 flags=b"\x00",
                 uid=b""):
        Request.__init__(self,
                         flags=flags,
                         command=b"\x02",
                         uid=uid)


class RequestReadSingleBlock(Request):

    def __init__(self,
                 flags=b"\x00",
                 uid_opt=b"",
                 block_nb=b""):
        Request.__init__(self,
                         flags=flags,
                         command=b"\x20",
                         uid_opt=uid_opt,
                         block_nb=block_nb)

    def resp_pretty_print(self, resp=b""):
        flags = resp[0:1]
        self.resp["flags"] = {"raw": flags,
                              "pretty": f"0x{flags[0]:02X}"}

        # Error flag?
        if resp[0] & 1:
            error_code = resp[1:2]
            self.resp["flags"] = {"raw": error_code,
                                  "pretty": f"{self.get_error_code(error_code[0])} ({error_code[0]:02X})"}
        else:
            if self.items["flags"][0] & 0x40:
                self.resp["block_security_status"] = resp[1:2]
                block_security_status = resp[1:2]
                self.resp["block_security_status"] = {"raw": block_security_status,
                                                      "pretty": f"{self.get_pretty_security_status(block_security_status)} ({block_security_status[0]:02X})"}
            self.resp["data"] = {"raw": resp[2:],
                                 "pretty": ""}


class RequestWriteSingleBlock(Request):

    def __init__(self,
                 flags=b"\x00",
                 uid_opt=b"",
                 block_nb=b"",
                 data=b""):
        Request.__init__(self,
                         flags=flags,
                         command=b"\x21",
                         uid_opt=uid_opt,
                         block_nb=block_nb,
                         data=data)


class RequestReadMultipleBlocks(Request):

    def __init__(self,
                 flags=b"\x00",
                 uid_opt=b"",
                 first_block_nb=b"\x00",
                 nb_blocks=b"\x00"):
        Request.__init__(self,
                         flags=flags,
                         command=b"\x23",
                         uid_opt=uid_opt,
                         first_block_nb=first_block_nb,
                         nb_blocks=nb_blocks)

        if uid_opt:
            self.items["uid_opt"] = uid_opt
        self.items["first_block_nb"] = first_block_nb
        self.items["nb_blocks"] = nb_blocks

    def resp_pretty_print(self, resp=b""):
        flags = resp[0:1]
        self.resp["flags"] = {"raw": flags,
                              "pretty": f"0x{flags[0]:02X}"}

        # Error flag?
        if resp[0] & 1:
            error_code = resp[1:2]
            self.resp["flags"] = {"raw": error_code,
                                  "pretty": f"{self.get_error_code(error_code[0])} ({error_code[0]:02X})"}
        else:
            nb_blocks = self.items["nb_blocks"][0] + 1
            chunk_size = len(resp[1:]) // nb_blocks
            if self.items["flags"][0] & 0x40:
                chunk_size -= 1
            cmpt = 2
            for hit in range(nb_blocks):
                if self.items["flags"][0] & 0x40:
                    block_security_status = resp[cmpt:cmpt + 1]
                    cmpt += 1
                    pretty = f"{self.get_pretty_security_status(block_security_status)} ({block_security_status[0]:02X})"
                    self.resp[f"block_security_status {hit}"] = {"raw": block_security_status,
                                                                 "pretty": pretty}
                data = resp[cmpt:cmpt + chunk_size]
                self.resp[f"block {hit}"] = {"raw": data,
                                             "pretty": b" ".join(f"{hit:02X}" for hit in data)}
                cmpt += chunk_size


class RequestWriteMultipleBlock(Request):

    def __init__(self,
                 flags=b"\x00",
                 uid_opt=b"",
                 first_block_nb=b"\x00",
                 nb_blocks=b"\x00",
                 data=b""):
        Request.__init__(self,
                         flags=flags,
                         command=b"\x24",
                         uid_opt=uid_opt,
                         first_block_nb=first_block_nb,
                         nb_blocks=nb_blocks,
                         data=data)


class RequestSelect(Request):

    def __init__(self,
                 flags=b"\x00",
                 uid=b""):
        Request.__init__(self,
                         flags=flags,
                         command=b"\x25",
                         uid=uid)


class RequestResetToReady(Request):

    def __init__(self,
                 flags=b"\x00",
                 uid_opt=b""):
        Request.__init__(self,
                         flags=flags,
                         command=b"\x25",
                         uid_opt=uid_opt)


class RequestGetMultipleBlockSecurityStatus(Request):

    def __init__(self,
                 flags=b"\x00",
                 uid_opt=b"",
                 first_block_nb=b"\x00",
                 nb_blocks=b"\x00"):
        Request.__init__(self,
                         flags=flags,
                         command=b"\x2C",
                         uid_opt=uid_opt,
                         first_block_nb=first_block_nb,
                         nb_blocks=nb_blocks)

    def resp_pretty_print(self, resp=b""):
        flags = resp[0:1]
        self.resp["flags"] = {"raw": flags,
                              "pretty": f"0x{flags[0]:02X}"}

        # Error flag?
        if resp[0] & 1:
            error_code = resp[1:2]
            self.resp["flags"] = {"raw": error_code,
                                  "pretty": f"{self.get_error_code(error_code[0])} ({error_code[0]:02X})"}
        else:
            nb = self.items["nb_blocks"][0] + 1
            for hit in range(1, len(resp)):
                block_security_status = resp[hit:hit + 1]
                pretty = f"{self.get_pretty_security_status(block_security_status)} ({block_security_status[0]:02X})"
                self.resp[f"block_security_status {nb}"] = {"raw": block_security_status,
                                                            "pretty": pretty}
                nb += 1


class RequestGetSystemInformation(Request):

    def __init__(self, flags=b"\x00", uid_opt=b""):
        Request.__init__(self,
                         flags=flags,
                         command=b"\x2B",
                         uid_opt=uid_opt[::-1])

        self.nb_block = None
        self.block_byte_size = None
        self.ic_reference = None

    def resp_pretty_print(self, resp=b""):
        flags = resp[0:1]
        self.resp["flags"] = {"raw": flags,
                              "pretty": f"0x{flags[0]:02X}"}

        # Error flag?
        if resp[0] & 1:
            error_code = resp[1:2]
            self.resp["flags"] = {"raw": error_code,
                                  "pretty": f"{self.get_error_code(error_code[0])} ({error_code[0]:02X})"}
        else:
            info_flags = resp[1:2]
            self.resp["info_flags"] = {"raw": info_flags,
                                       "pretty": self.pretty_print_request_info_flags(info_flags[0])}

            uid = resp[2:10]
            self.resp["uid"] = {"raw": uid,
                                "pretty": " ".join(f"{hit:02X}" for hit in resp[2:10][::-1])}

            cmpt = 10

            if resp[1] & 1:
                dsfid = resp[cmpt:cmpt + 1]
                self.resp["dsfid"] = {"raw": dsfid,
                                      "pretty": " ".join(f"{hit:02X}" for hit in dsfid)}
                cmpt += 1

            if resp[1] & 2:
                afi = resp[cmpt:cmpt + 1]
                self.resp["afi"] = {"raw": afi,
                                    "pretty": " ".join(f"{hit:02X}" for hit in dsfid)}
                cmpt += 1

            if resp[1] & 4:
                vicc_memory_size = resp[cmpt:cmpt + 2]
                self.nb_block = resp[cmpt] + 1
                cmpt += 1
                self.block_byte_size = (resp[cmpt] & 0x1F) + 1
                cmpt += 1

                self.resp["vicc_memory_size"] = {"raw": vicc_memory_size,
                                                 "pretty": f"{self.nb_block} blocks of {self.block_byte_size} bytes"}



            if resp[1] & 8:
                ic_reference = resp[cmpt:cmpt + 1]
                try:
                    ic_reference_pretty = manufacturer_codes_iso_7816_6(ic_reference.hex().upper())
                except:  # noqa: E722
                    ic_reference_pretty = "Unknown code"
                self.resp["ic_reference"] = {"raw": ic_reference,
                                             "pretty": ic_reference_pretty}



# Copyright (C) 2015-2020 Guillaume VINET
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Tpdu(object):

    def __init__(self, tpdu):
        self._tpdu = tpdu
        self._pcb, self._nad, self._cid, self._inf_field, self._crc = None, None, None, None, None
        self.iblock_is_chaining, self.is_nad_present, self.is_cid_present = False, False, False
        self.parse_block()

    def parse_pcb(self):
        pcb = self._pcb
        # Iblock
        if (pcb & 0xC0) == 0x00:
            self.iblock_is_chaining = ((pcb & 0x10) == 0x10)
            self.is_nad_present = (pcb & 0x40) == 0x40
            self.is_cid_present = (pcb & 0x08) == 0x08
        # Rblock
        elif (pcb & 0xC0) == 0x80:
            self.is_cid_present = (pcb & 0x08) == 0x08
            self.is_nad_present = False
        # Sblock
        elif (pcb & 0xC0) == 0xC0:
            self.is_cid_present = (pcb & 0x08) == 0x08
            self.is_nad_present = False

    def parse_block(self):
        self._pcb = self._tpdu[0]
        self.parse_pcb()
        cmpt = 1
        if self.is_cid_present:
            self._cid = self._tpdu[cmpt]
            cmpt += 1

        if self.is_nad_present:
            self._nad = self._tpdu[cmpt]
            cmpt += 1

        self._inf_field = self._tpdu[cmpt:-2]
        self._crc = self._tpdu[-2:]

    def get_tpdu(self):
        return self._tpdu

    def get_inf_field(self):
        return self._inf_field

    def is_chaining(self):
        return self.iblock_is_chaining

    def is_wtx(self):
        return (self._pcb & 0xF0) == 0xF0

    def get_wtx_reply(self):

        resp = [self._pcb]

        if self.is_cid_present:
            resp.append(self._cid)

        if self.is_nad_present:
            resp.append(self._nad)

        # inf field
        resp.append(self._inf_field[0] & 0x3F)

        return resp



# Copyright (C) 2020 Guillaume VINET
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pynfcreader
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

name = 'pyNFCReader'
version = str(pynfcreader.__version__)
release = str(version)

setuptools.setup(
    name=name,
    version=pynfcreader.__version__,
    author="Guillaume VINET",
    description="NFC reader",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gvinet/pynfcreader",
    packages=setuptools.find_packages(),
    install_requires=['pyHydrabus'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
